/****************************************************************************
*
*  LineF.c
*
*  Implements the line follower logic
*
****************************************************************************/

#pragma chip PIC16F877

// ---- Include Files -------------------------------------------------------

#define  DEBUG_LF 1

#define  DEBUG DEBUG_LF

#include "DebugKey.h"
#include "LineF.h"
#include "Motor.h"
#include "Delay.h"
#include "Puts.h"
#include "Sensor.h"

#if DEBUG_LF
#define  DBG_LF(x)      x
#else
#define  DBG_LF(x)
#endif

// ---- Public Variables ----------------------------------------------------
// ---- Private Constants and Types -----------------------------------------
// ---- Private Variables ---------------------------------------------------

#define  LINE_STATE_L   0x01
#define  LINE_STATE_CL  0x02
#define  LINE_STATE_C   0x04
#define  LINE_STATE_CR  0x08
#define  LINE_STATE_R   0x10

#define  LINE_THRESH    200

#define  LINE_HINT_LEFT    1
#define  LINE_HINT_RIGHT   2

// The line seen timer tracks time in 76Hz ticks.

#define  LINE_SEEN_MAX_TIME     255
#define  LINE_SEEN_THRESH_TIME  38  // 1/2 second

uns8  gNoLineCounter;
uns8  gLineHint;

uns8  gLineSeenR;
uns8  gLineSeenL;
uns8  gLineSeenTime;

uns8  gLineState;
uns8  gPrevLineState;

// ---- Private Function Prototypes -----------------------------------------
// ---- Functions -----------------------------------------------------------

/****************************************************************************
*
* Pulse the motors, but advance our timers at the same time.
*/

void LinePulseMotors( void )
{
#if DEBUG_LF
    if (( gSpeedL != gPrevSpeedL ) || ( gSpeedR != gPrevSpeedR ))
    {
       puts( "PulseMotors L:" );
       putu8( gSpeedL );
       puts( " R:" );
       putu8( gSpeedR );
       puts( " t0:" );
       putu8( gT0 );
       puts( " t1:" );
       putu8( gT1 );
       puts( " C: " );
       putu16( gHuntCounter );
       puts( "\n" );

       gPrevSpeedL = gSpeedL;
       gPrevSpeedR = gSpeedR;
    }
#endif

    PulseMotors();

    if( gLineSeenTime < LINE_SEEN_MAX_TIME )
    {
        gLineSeenTime++;
    }

} // LinePulseMotors

/****************************************************************************
*
*  Do the line following
*/

extern void LineFollower( void )
{
    uns8  i;
    uns8  timeToReacquire;

    gSpeedL = SPEED_FWD;
    gSpeedR = SPEED_FWD;
    gNoLineCounter = 0;
    DBG_LF( gHuntCounter = 0; )

    gLineSeenL = 0;
    gLineSeenR = 0;
    gLineSeenTime = LINE_SEEN_MAX_TIME;

    gPrevLineState = 0;
    gLineState = 0;

    gRampDelta = 8;

    DBG_LF( puts( "Line Follower started\n" ); )

    while ( 1 )
    {
        // Wait for 76 Hz timer

        WaitForTimer0Rollover();

#if DEBUG_LF
        if ( RCIF == 1 )
        {
            uns8  ch;
            ch = RCREG;

            DebugKey( ch );
        }
#endif

        ReadSensors();

        gLineState = 0;
        if ( gLineL > LINE_THRESH )
        {
            gLineState |= LINE_STATE_L;
        }
        if ( gLineCL > LINE_THRESH )
        {
            gLineState |= LINE_STATE_CL;
        }
        if ( gLineC > LINE_THRESH )
        {
            gLineState |= LINE_STATE_C;
        }
        if ( gLineCR > LINE_THRESH )
        {
            gLineState |= LINE_STATE_CR;
        }
        if ( gLineR > LINE_THRESH )
        {
            gLineState |= LINE_STATE_R;
        }

#if DEBUG_LF
        if ( gLineState != gPrevLineState )
        {
            puts( "L[ " );
            putu8( gLineL );
            puts( " " );                        
            putu8( gLineCL );
            puts( " " );
            putu8( gLineC );
            puts( " " );
            putu8( gLineCR );
            puts( " " );
            putu8( gLineR );
            puts( " ]R Seen[ " );
            putdigit( gLineSeenL );
            putdigit( gLineSeenR );
            puts( ":" );
            putu8( gLineSeenTime );
            puts( " ]\n" );
        }
#endif

        if (( gLineR > LINE_THRESH ) || ( gLineL > LINE_THRESH ))
        {
            if ( gLineSeenTime > LINE_SEEN_THRESH_TIME )
            {
                // It's been a while since we last saw a line transition.
                // Assume that this is a "new" transition.

                gLineSeenL = 0;
                gLineSeenR = 0;

                DBG_LF( puts( "Reseting Line Seen - new transition\n" ); )
            }

            if ( gLineL > LINE_THRESH )
            {
                gLineSeenL = 1;
            }

            if ( gLineR > LINE_THRESH )
            {
                gLineSeenR = 1;
            }
            gLineSeenTime = 0;
        }
        
        if ( gLineC > LINE_THRESH )
        {
            // Center sensor is activated - We keep going straight regardless
            // of what the other sensors say

            gSpeedL = SPEED_FWD;
            gSpeedR = SPEED_FWD;
        }
        else
        if ( gLineCL > LINE_THRESH )
        {
            // Sensor left of center is activated

            gSpeedL = SPEED_FWD_MED;
            gSpeedR = SPEED_FWD;
        }
        else
        if ( gLineCR > LINE_THRESH )
        {
            // Sensor right of center is activated

            gSpeedL = SPEED_FWD;
            gSpeedR = SPEED_FWD_MED;
        }
        else
        if ( gLineL > LINE_THRESH )
        {
            // Far left sensor is activated

            gSpeedL = SPEED_OFF;
            gSpeedR = SPEED_FWD;
        }
        else
        if ( gLineR > LINE_THRESH )
        {
            // Far right sensor is activated

            gSpeedL = SPEED_FWD;
            gSpeedR = SPEED_OFF;
        }
        else
        {
            // No sensors are active - We just ran out of line.

            // If the line was seen on both sides (or neither side)
            // then we want to keep going forward).

            if ( gLineSeenL != gLineSeenR )
            {
#if 1
                // Stop for a bit

                gSpeedL = SPEED_OFF;
                gSpeedR = SPEED_OFF;

                for ( i = 0; i < 5; i++ )
                {
                    LinePulseMotors();
                    WaitForTimer0Rollover();
                    ReadSensors();
                }
#endif

                if ( gLineSeenL )
                {
                    // Line seen on the left. Turn left until the center sensor reacquires

                    gSpeedL = SPEED_BWD_SPIN;
                    gSpeedR = SPEED_FWD_SPIN;
                }
                else
                {
                    // Line seen on the right. Turn right until the center sensor reacquires

                    gSpeedL = SPEED_FWD_SPIN;
                    gSpeedR = SPEED_BWD_SPIN;
                }

                // Turn until the center sensor reacquires

                timeToReacquire = 0;
                while ( gLineC <= LINE_THRESH )
                {
                    LinePulseMotors();
                    WaitForTimer0Rollover();
                    ReadSensors();
                    if( timeToReacquire < 255 )
                    {
                        timeToReacquire++;
                    }
                }

                // If this was a "major" correction (i.e. we turned 90 degrees)
                // then we want to reset the line seen indicators, since we
                // want to retrigger on new dections.

                // However, if it was a "minor" correction (we just happened to
                // run off the edge of the line and we quickly reacquired the
                // line again, then we don't want to forget about the recently
                // seen line transition.

                DBG_LF( puts( "Time to reacquire line: " ); )
                DBG_LF( putu8( timeToReacquire ); )
                DBG_LF( puts( "\n" ); )

                // With HiTec motors, it only takes about a count of 17 to 
                // turn 90 degrees.
                
                if ( timeToReacquire > 10 ) // 36 = 1/2 second
                {
                    gLineSeenL = 0;
                    gLineSeenR = 0;

                    DBG_LF( puts( "Reseting Line Seen - reacquire\n" ); )
                }
            }
        }

        LinePulseMotors();

        DBG_LF( gDebugCounter++; )
        gPrevLineState = gLineState;
    }

} // LineFollower

